# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""The ``init`` command creates a new project."""
from __future__ import absolute_import, print_function

import os

from anaconda_project import project_ops
from anaconda_project.internal.cli.console_utils import (print_project_problems, console_ask_yes_or_no)


def init_command(project_dir, assume_yes, with_anaconda_package):
    """Initialize a new project.

    Returns:
        Exit code (0 on success)
    """
    # we don't want False right now because either you specify
    # --yes or we go with the default in project_ops.create
    # (depends on whether project file already exists).
    assert assume_yes is None or assume_yes is True
    assert with_anaconda_package is None or with_anaconda_package is True

    if not os.path.exists(project_dir):
        if assume_yes:
            make_directory = True
        else:
            make_directory = console_ask_yes_or_no("Create directory '%s'?" % project_dir, default=False)
    else:
        make_directory = False

    project = project_ops.create(project_dir,
                                 make_directory=make_directory,
                                 fix_problems=assume_yes,
                                 with_anaconda_package=with_anaconda_package)
    if print_project_problems(project):
        return 1
    else:
        print("Project configuration is in %s" % project.project_file.filename)
        return 0


def main(args):
    """Start the init command and return exit status code."""
    return init_command(args.directory, args.yes, args.with_anaconda_package)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import platform
import pytest

from anaconda_project.internal.windows_cmdline import (windows_split_command_line, windows_join_command_line,
                                                       WindowsCommandLineException)


def _roundtrip(args):
    if platform.system() != 'Windows':
        return
    joined = windows_join_command_line(args)
    assert isinstance(joined, str)
    split = windows_split_command_line(joined)
    assert isinstance(split, list)
    for s in split:
        assert isinstance(s, str)
    assert split == args


def test_empty_args():
    with pytest.raises(WindowsCommandLineException) as excinfo:
        windows_join_command_line([])
    assert "Windows has no way to encode an empty arg list as a command line" in str(excinfo.value)


def test_single_arg():
    _roundtrip(["foo"])


def test_single_arg_with_spaces():
    _roundtrip(["foo bar"])


def test_single_arg_with_backslash():
    _roundtrip(["foo\\bar"])


def test_single_arg_with_double_quote():
    _roundtrip(['foo"bar'])


def test_single_arg_which_is_a_space():
    _roundtrip([" "])


def test_multiple_args():
    _roundtrip(["foo", "bar"])


def test_multiple_args_with_spaces_and_quotes():
    _roundtrip(["foo", "bl ah", " bar ", "b\"az"])


def test_multiple_args_with_backslash():
    _roundtrip(["foo", "bl\\ah"])


def test_multiple_args_with_unicode():
    try:
        _roundtrip(["foo", "bÄr", "¿"])
    except WindowsCommandLineException as e:
        # on Win32/py2 apparently we end up in ASCII
        if 'cannot represent this command line in its character encoding' not in str(e):
            raise e


def test_multiple_args_which_are_spaces():
    _roundtrip([" ", " ", " "])


def test_cannot_join_first_arg_starting_with_quote():
    with pytest.raises(WindowsCommandLineException) as excinfo:
        windows_join_command_line(['"foo'])
    assert 'Windows does not allow the first arg to start with a quote' in str(excinfo.value)


def test_cannot_join_first_arg_with_quote_and_space():
    with pytest.raises(WindowsCommandLineException) as excinfo:
        windows_join_command_line(['foo"bar baz'])
    assert 'Windows does not allow the first arg to have both quotes and spaces' in str(excinfo.value)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

from anaconda_project.internal.http_client import FileDownloader
from anaconda_project.internal.test.http_server import HttpServerTestContext
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents

from tornado.ioloop import IOLoop

import os
import sys
import platform
import stat


def _download_file(length, hash_algorithm):
    def inside_directory_download_file(dirname):
        filename = os.path.join(dirname, "downloaded-file")
        with HttpServerTestContext() as server:
            url = server.new_download_url(download_length=length, hash_algorithm=hash_algorithm)
            download = FileDownloader(url=url, filename=filename, hash_algorithm=hash_algorithm)
            response = IOLoop.current().run_sync(download.run)
            assert [] == download.errors
            assert response is not None
            assert response.code == 200
            if hash_algorithm:
                server_hash = server.server_computed_hash_for_downloaded_url(url)
                assert download.hash == server_hash
            statinfo = os.stat(filename)
            assert statinfo.st_size == length
            assert not os.path.isfile(filename + ".part")

    with_directory_contents(dict(), inside_directory_download_file)


def test_download_empty_file_md5():
    _download_file(0, 'md5')


def test_download_small_file_md5():
    _download_file(1024, 'md5')


def test_download_small_file_hashless():
    _download_file(1024, None)


def test_download_medium_file_md5():
    _download_file(1024 * 1024, 'md5')


def test_download_medium_file_sha1():
    _download_file(1024 * 1024, 'sha1')


# this takes too long so disabled via underscore-prefix by default.
# uncomment it for manual testing if desired.
def _test_download_huge_file_md5():
    kilo = 1024
    mega = kilo * 1024
    giga = mega * 1024
    _download_file(int(giga * 0.2), 'md5')


def test_download_has_http_error():
    def inside_directory_get_http_error(dirname):
        filename = os.path.join(dirname, "downloaded-file")
        with HttpServerTestContext() as server:
            url = server.error_url
            download = FileDownloader(url=url, filename=filename, hash_algorithm='md5')
            response = IOLoop.current().run_sync(download.run)
            assert ['Failed download to %s: HTTP 404: Not Found' % filename] == download.errors
            assert response is None
            assert not os.path.isfile(filename)
            assert not os.path.isfile(filename + ".part")

    with_directory_contents(dict(), inside_directory_get_http_error)


def test_download_fail_to_create_directory(monkeypatch):
    def inside_directory_fail_to_create_directory(dirname):
        def mock_makedirs(name):
            raise IOError("Cannot create %s" % name)

        monkeypatch.setattr('anaconda_project.internal.makedirs.makedirs_ok_if_exists', mock_makedirs)
        filename = os.path.join(dirname, "downloaded-file")
        with HttpServerTestContext() as server:
            url = server.error_url
            download = FileDownloader(url=url, filename=filename, hash_algorithm='md5')
            response = IOLoop.current().run_sync(download.run)
            assert ["Could not create directory '%s': Cannot create %s" % (dirname, dirname)] == download.errors
            assert response is None
            assert not os.path.isfile(filename)
            assert not os.path.isfile(filename + ".part")

    with_directory_contents(dict(), inside_directory_fail_to_create_directory)


def test_download_fail_to_open_file(monkeypatch):
    def inside_directory_fail_to_open_file(dirname):
        statinfo = os.stat(dirname)
        try:
            filename = os.path.join(dirname, "downloaded-file")
            # make the open fail
            if platform.system() == 'Windows':
                # windows does not have read-only directories so we have to
                # make the file read-only
                with open(filename + ".part", 'wb') as file:
                    file.write("".encode())
                os.chmod(filename + ".part", stat.S_IREAD)
            else:
                os.chmod(dirname, stat.S_IREAD)

            with HttpServerTestContext() as server:
                url = server.error_url
                download = FileDownloader(url=url, filename=filename, hash_algorithm='md5')
                response = IOLoop.current().run_sync(download.run)
                filename_with_weird_extra_slashes = filename
                if platform.system() == 'Windows':
                    # I dunno. that's what Windows gives us.
                    filename_with_weird_extra_slashes = filename.replace("\\", "\\\\")
                assert [
                    "Failed to open %s.part: [Errno 13] Permission denied: '%s.part'" %
                    (filename, filename_with_weird_extra_slashes)
                ] == download.errors
                assert response is None
                assert not os.path.isfile(filename)
                if platform.system() != 'Windows':
                    # on windows we created this ourselves to cause the open error above
                    assert not os.path.isfile(filename + ".part")
        finally:
            # put this back so we don't get an exception cleaning up the directory
            os.chmod(dirname, statinfo.st_mode)
            if platform.system() == 'Windows':
                os.chmod(filename + ".part", stat.S_IWRITE)

    with_directory_contents(dict(), inside_directory_fail_to_open_file)


class _FakeFileFailsToWrite(object):
    def write(self, chunk):
        raise IOError("FAIL")

    def close(self):
        pass


def test_download_fail_to_write_file(monkeypatch):
    def inside_directory_fail_to_write_file(dirname):
        filename = os.path.join(dirname, "downloaded-file")
        with HttpServerTestContext() as server:
            url = server.new_download_url(download_length=(1024 * 1025), hash_algorithm='md5')
            download = FileDownloader(url=url, filename=filename, hash_algorithm='md5')

            def mock_open(filename, mode):
                return _FakeFileFailsToWrite()

            if sys.version_info > (3, 0):
                monkeypatch.setattr('builtins.open', mock_open)
            else:
                monkeypatch.setattr('__builtin__.open', mock_open)

            response = IOLoop.current().run_sync(download.run)
            assert ["Failed to write to %s: FAIL" % (filename + ".part")] == download.errors
            assert response.code == 200
            assert not os.path.isfile(filename)
            assert not os.path.isfile(filename + ".part")

    with_directory_contents(dict(), inside_directory_fail_to_write_file)


def test_download_fail_to_rename_tmp_file(monkeypatch):
    def inside_directory_fail_to_rename_tmp_file(dirname):
        filename = os.path.join(dirname, "downloaded-file")
        with HttpServerTestContext() as server:
            url = server.new_download_url(download_length=56780, hash_algorithm='md5')
            download = FileDownloader(url=url, filename=filename, hash_algorithm='md5')

            def mock_rename(src, dest):
                raise OSError("FAIL")

            monkeypatch.setattr('anaconda_project.internal.rename.rename_over_existing', mock_rename)

            response = IOLoop.current().run_sync(download.run)
            assert ["Failed to rename %s to %s: FAIL" % (filename + ".part", filename)] == download.errors
            assert response.code == 200
            assert not os.path.isfile(filename)
            assert not os.path.isfile(filename + ".part")

    with_directory_contents(dict(), inside_directory_fail_to_rename_tmp_file)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import print_function, absolute_import

import codecs
import os
import sys
try:
    from backports.tempfile import TemporaryDirectory
except ImportError:
    from tempfile import TemporaryDirectory
import zipfile
import tempfile

from anaconda_project.internal.makedirs import makedirs_ok_if_exists
from anaconda_project.local_state_file import LocalStateFile
from anaconda_project.yaml_file import _load_string
from anaconda_project.project_file import (possible_project_file_names, DEFAULT_PROJECT_FILENAME)

local_tmp = os.path.abspath("./build/tmp")
makedirs_ok_if_exists(local_tmp)


def with_directory_contents(contents, func):
    tempd = TemporaryDirectory(prefix="test-")
    dirname = tempd.name
    try:
        for filename, file_content in contents.items():
            path = os.path.join(dirname, filename)
            if file_content is None:
                # make a directory
                makedirs_ok_if_exists(path)
            else:
                makedirs_ok_if_exists(os.path.dirname(path))
                with codecs.open(path, 'w', 'utf-8') as f:
                    f.write(file_content)
        result = func(os.path.realpath(dirname))
    finally:
        # Windows experiences PermissionError exceptions here,
        # and Unix sometimes experiences FileNotFound exceptions.
        # The reasons are not 100% clear, but they should not be
        # allowed to interrupt test passage, either.
        try:
            tempd.cleanup()
        except Exception as exc:
            print('Unexpected error cleaning temporary directory:')
            print('  ' + dirname)
            print('  ' + str(exc))
            pass
    return result


def complete_project_file_content(content):
    yaml = _load_string(content)
    if yaml is None:
        raise AssertionError("Broken yaml: %r" % content)

    modified = content
    if 'env_specs' not in yaml:
        modified = (modified + "\n" + "env_specs:\n" + "  default:\n" + "    description: default\n" + "\n")

    if 'name' not in yaml:
        modified = (modified + "\n" + "name: some_name\n")

    if 'platforms' not in yaml:
        modified = (modified + "\n" + "platforms: [linux-64, osx-64, win-64]\n")

    if modified is not content:
        try:
            # make sure we didn't mangle it
            _load_string(modified)
            return modified
        except Exception as e:
            print("Failed to parse: " + modified, file=sys.stderr)
            raise e
    else:
        return content


def with_directory_contents_completing_project_file(contents, func):
    new_contents = {}
    for filename, file_content in contents.items():
        if filename in possible_project_file_names:
            file_content = complete_project_file_content(file_content)
        new_contents[filename] = file_content
    if len([key for key in new_contents.keys() if key in possible_project_file_names]) == 0:
        new_contents[DEFAULT_PROJECT_FILENAME] = complete_project_file_content("")
    return with_directory_contents(new_contents, func)


def with_temporary_file(func, dir=None):
    if dir is None:
        dir = local_tmp
    # Windows throws a permission denied if we use delete=True for
    # auto-delete, and then try to open the file again ourselves
    # with f.name. So we manually delete in the finally block
    # below.
    f = tempfile.NamedTemporaryFile(dir=dir, delete=False)
    try:
        return func(f)
    finally:
        f.close()
        os.remove(f.name)


def with_named_file_contents(filename, contents, func, dir=None):
    if dir is None:
        dir = local_tmp

    with TemporaryDirectory(prefix="test-") as dirname:
        full = os.path.join(dirname, filename)
        with codecs.open(full, 'w', encoding='utf-8') as f:
            f.write(contents)
            f.flush()
        return func(full)


def with_file_contents(contents, func, dir=None):
    def with_file_object(f):
        f.write(contents.encode("UTF-8"))
        f.flush()
        # Windows will get mad if we try to rename it without closing,
        # and some users of with_file_contents want to rename it.
        f.close()
        return func(f.name)

    return with_temporary_file(with_file_object, dir=dir)


def with_temporary_script_commandline(contents, func, dir=None):
    def script_wrapper(filename):
        return func(['python', os.path.abspath(filename)])

    return with_file_contents(contents, script_wrapper, dir=dir)


def tmp_script_commandline(contents):
    import tempfile
    # delete=False required so windows will allow the file to be
    # opened.  (we never delete this tmpfile, except when the
    # entire tmpdir is blown away)
    f = tempfile.NamedTemporaryFile(dir=local_tmp, delete=False, suffix=".py", prefix="script_")
    f.write(contents.encode('utf-8'))
    f.close()
    return ['python', os.path.abspath(f.name)]


def tmp_local_state_file():
    import tempfile
    # delete=False required so windows will allow the file to be opened
    f = tempfile.NamedTemporaryFile(dir=local_tmp, delete=False)
    local_state = LocalStateFile(f.name)
    f.close()
    os.remove(f.name)
    return local_state


def with_tmp_zipfile(contents, f):
    """Call 'f' with a zip of 'contents' and an empty working directory name."""
    def using_temporary_file(handle):
        with zipfile.ZipFile(handle.name, 'w') as zf:
            for key, value in contents.items():
                zf.writestr(key, value.encode('utf-8'))

        def using_directory(dirname):
            return f(handle.name, dirname)

        return with_directory_contents(dict(), using_directory)

    return with_temporary_file(using_temporary_file)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
import codecs
import os

from anaconda_project.internal.test.tmpfile_utils import with_directory_contents
from anaconda_project.test.project_utils import assert_identical_except_blank_lines
from anaconda_project.project_lock_file import (ProjectLockFile, DEFAULT_PROJECT_LOCK_FILENAME,
                                                possible_project_lock_file_names)
from anaconda_project.conda_manager import CondaLockSet

expected_default_file = """# This is an Anaconda project lock file.
# The lock file locks down exact versions of all your dependencies.
#
# In most cases, this file is automatically maintained by the `anaconda-project` command or GUI tools.
# It's best to keep this file in revision control (such as git or svn).
# The file is in YAML format, please see http://www.yaml.org/start.html for more.
#

#
# Set to false to ignore locked versions.
#
locking_enabled: false

#
# A key goes in here for each env spec.
#
env_specs: {}
"""


def _get_locking_enabled(lock_file, env_spec_name):
    """Library-internal method."""
    enabled = lock_file.get_value(['env_specs', env_spec_name, 'locked'], None)
    if enabled is not None:
        return enabled

    enabled = lock_file.get_value(['locking_enabled'])
    if enabled is not None:
        return enabled

    return True


def _get_lock_set(lock_file, env_spec_name):
    """Library-internal method."""
    # TODO no validation here, we'll do that by moving this
    # into project.py soon
    enabled = _get_locking_enabled(lock_file, env_spec_name)
    packages = lock_file.get_value(['env_specs', env_spec_name, 'packages'], {})
    platforms = lock_file.get_value(['env_specs', env_spec_name, 'platforms'], [])
    env_spec_hash = lock_file.get_value(['env_specs', env_spec_name, 'env_spec_hash'], None)
    lock_set = CondaLockSet(packages, platforms, enabled=enabled)
    lock_set.env_spec_hash = env_spec_hash
    return lock_set


def test_create_missing_lock_file_only_when_not_default():
    def create_file(dirname):
        filename = os.path.join(dirname, DEFAULT_PROJECT_LOCK_FILENAME)
        assert not os.path.exists(filename)
        lock_file = ProjectLockFile.load_for_directory(dirname)
        assert lock_file is not None
        assert not os.path.exists(filename)

        assert _get_lock_set(lock_file, 'foo').disabled
        assert not _get_locking_enabled(lock_file, 'foo')

        lock_file.save()
        # should not have saved an unmodified (default) file)
        assert not os.path.exists(filename)

        # make a change, which should cause us to save
        lock_file.set_value(['something'], 42)

        lock_file.save()
        assert os.path.exists(filename)
        with codecs.open(filename, 'r', 'utf-8') as file:
            contents = file.read()
            expected = expected_default_file + "something: 42\n"
            assert_identical_except_blank_lines(expected, contents)

    with_directory_contents(dict(), create_file)


def _use_existing_lock_file(relative_name):
    def check_file(dirname):
        filename = os.path.join(dirname, relative_name)
        assert os.path.exists(filename)
        lock_file = ProjectLockFile.load_for_directory(dirname)
        assert lock_file.get_value(['env_specs', 'foo']) is not None
        assert lock_file.get_value(['locking_enabled']) is True

    with_directory_contents(
        {
            relative_name:
            """
locking_enabled: true
env_specs:
  foo:
    locked: true
    platforms: [linux-32,linux-64,osx-64,win-32,win-64]
    packages:
      all:
        - foo=1.0=1
  bar:
    locked: false
    platforms: [linux-32,linux-64,osx-64,win-32,win-64]
    packages:
      all:
        - bar=2.0=2
"""
        }, check_file)


def test_use_existing_lock_file_default_name():
    _use_existing_lock_file(DEFAULT_PROJECT_LOCK_FILENAME)


def test_use_existing_lock_file_all_names():
    for name in possible_project_lock_file_names:
        _use_existing_lock_file(name)


def _use_existing_lock_file_from_subdir(relative_name):
    def check_file(dirname):
        filename = os.path.join(dirname, relative_name)
        assert os.path.exists(filename)
        subdir = os.path.join(dirname, 'subdir')
        os.makedirs(subdir)
        lock_file = ProjectLockFile.load_for_directory(subdir)
        assert lock_file.get_value(['env_specs', 'foo']) is not None
        assert lock_file.get_value(['locking_enabled']) is True

    with_directory_contents(
        {
            relative_name:
            """
locking_enabled: true
env_specs:
  foo:
    locked: true
    platforms: [linux-32,linux-64,osx-64,win-32,win-64]
    packages:
      all:
        - foo=1.0=1
  bar:
    locked: false
    platforms: [linux-32,linux-64,osx-64,win-32,win-64]
    packages:
      all:
        - bar=2.0=2
"""
        }, check_file)


def test_use_existing_lock_file_default_name_from_subdir():
    _use_existing_lock_file_from_subdir(DEFAULT_PROJECT_LOCK_FILENAME)


def test_use_existing_lock_file_all_names_from_subdir():
    for name in possible_project_lock_file_names:
        _use_existing_lock_file_from_subdir(name)


def test_get_lock_set():
    def check_file(dirname):
        filename = os.path.join(dirname, DEFAULT_PROJECT_LOCK_FILENAME)
        assert os.path.exists(filename)
        lock_file = ProjectLockFile.load_for_directory(dirname)
        foo_lock_set = _get_lock_set(lock_file, 'foo')
        assert foo_lock_set.enabled
        assert ('foo=1.0=1', ) == foo_lock_set.package_specs_for_current_platform
        assert ['qbert==1.0.0'] == foo_lock_set.pip_package_specs
        bar_lock_set = _get_lock_set(lock_file, 'bar')
        assert bar_lock_set.disabled

    with_directory_contents(
        {
            DEFAULT_PROJECT_LOCK_FILENAME:
            """
locking_enabled: true
env_specs:
  foo:
    locked: true
    platforms: [linux-32,linux-64,osx-64,win-32,win-64]
    packages:
      all:
        - foo=1.0=1
      pip:
        - qbert==1.0.0
  bar:
    locked: false
    platforms: [linux-32,linux-64,osx-64,win-32,win-64]
    packages:
      all:
        - bar=2.0=2
      pip:
        - pbert==1.1.0
"""
        }, check_file)


def test_disable_single_spec_locking():
    def check_file(dirname):
        filename = os.path.join(dirname, DEFAULT_PROJECT_LOCK_FILENAME)
        assert os.path.exists(filename)
        lock_file = ProjectLockFile.load_for_directory(dirname)
        foo_lock_set = _get_lock_set(lock_file, 'foo')
        assert ('foo=1.0=1', ) == foo_lock_set.package_specs_for_current_platform

        lock_file._disable_locking('foo')

        foo_lock_set = _get_lock_set(lock_file, 'foo')
        assert foo_lock_set.disabled
        assert _get_locking_enabled(lock_file, 'foo') is False

    with_directory_contents(
        {
            DEFAULT_PROJECT_LOCK_FILENAME:
            """
locking_enabled: true
env_specs:
  foo:
    locked: true
    platforms: [linux-32,linux-64,osx-64,win-32,win-64]
    packages:
      all:
        - foo=1.0=1
  bar:
    locked: false
    platforms: [linux-32,linux-64,osx-64,win-32,win-64]
    packages:
      all:
        - bar=2.0=2
"""
        }, check_file)


def test_set_lock_set():
    def check_file(dirname):
        filename = os.path.join(dirname, DEFAULT_PROJECT_LOCK_FILENAME)
        assert os.path.exists(filename)
        lock_file = ProjectLockFile.load_for_directory(dirname)

        # we have the global enabled flag off; individual
        # env spec settings override that; foo has no setting.
        foo_lock_set = _get_lock_set(lock_file, 'foo')
        assert lock_file.get_value(['env_specs', 'foo', 'locked'], None) is None
        assert foo_lock_set.disabled
        bar_lock_set = _get_lock_set(lock_file, 'bar')
        assert bar_lock_set.disabled

        all_names = ['foo', 'bar']

        lock_set = CondaLockSet({'all': ['something=3.0=0']},
                                platforms=['linux-32', 'linux-64', 'osx-64', 'win-32', 'win-64'])
        lock_set.env_spec_hash = "hash-hash-hash"

        lock_file._set_lock_set('bar', lock_set, all_names=all_names)

        # "foo" should have been DISABLED since we had to
        # enable the global flag in order to enable "bar"
        foo_lock_set = _get_lock_set(lock_file, 'foo')
        assert lock_file.get_value(['env_specs', 'foo', 'locked']) is False
        assert foo_lock_set.disabled
        assert foo_lock_set.env_spec_hash is None

        bar_lock_set = _get_lock_set(lock_file, 'bar')
        assert bar_lock_set.enabled
        assert ('something=3.0=0', ) == bar_lock_set.package_specs_for_current_platform
        assert "hash-hash-hash" == bar_lock_set.env_spec_hash

        # and now we should enable "foo" when we set it to something
        lock_file._set_lock_set('foo', lock_set, all_names=all_names)
        foo_lock_set = _get_lock_set(lock_file, 'foo')
        assert foo_lock_set.enabled
        assert ('something=3.0=0', ) == foo_lock_set.package_specs_for_current_platform

        # be sure we can save
        lock_file.save()

        reloaded = ProjectLockFile.load_for_directory(dirname)

        assert ('something=3.0=0', ) == _get_lock_set(reloaded, 'bar').package_specs_for_current_platform
        assert ('something=3.0=0', ) == _get_lock_set(reloaded, 'foo').package_specs_for_current_platform

        # Check _set_lock_set_hash
        lock_file._set_lock_set_hash('bar', 'hash2.0')
        lock_file.save()

        reloaded = ProjectLockFile.load_for_directory(dirname)
        bar_lock_set = _get_lock_set(reloaded, 'bar')
        assert bar_lock_set.env_spec_hash == 'hash2.0'

    with_directory_contents(
        {
            DEFAULT_PROJECT_LOCK_FILENAME:
            """
locking_enabled: false
env_specs:
  foo:
    platforms: [linux-32,linux-64,osx-64,win-32,win-64]
    packages:
      all:
        - foo=1.0=1
  bar:
    locked: false
    platforms: [linux-32,linux-64,osx-64,win-32,win-64]
    packages:
      all:
        - bar=2.0=2
"""
        }, check_file)


def test_set_lock_set_has_to_create_env_specs_to_disable():
    def check_file(dirname):
        filename = os.path.join(dirname, DEFAULT_PROJECT_LOCK_FILENAME)
        assert os.path.exists(filename)
        lock_file = ProjectLockFile.load_for_directory(dirname)

        all_names = ['foo', 'bar']

        lock_set = CondaLockSet({'all': ['something=3.0=0']},
                                platforms=['linux-32', 'linux-64', 'osx-64', 'win-32', 'win-64'])

        # so the point of this test is that we need to create env_specs
        # dict and the 'foo' entry as a side effect of setting 'bar',
        # in order to mark 'foo' disabled.
        lock_file._set_lock_set('bar', lock_set, all_names=all_names)

        # "foo" should have been DISABLED since we had to
        # enable the global flag in order to enable "bar"
        foo_lock_set = _get_lock_set(lock_file, 'foo')
        assert lock_file.get_value(['env_specs', 'foo', 'locked']) is False
        assert foo_lock_set.disabled

        bar_lock_set = _get_lock_set(lock_file, 'bar')
        assert bar_lock_set.enabled
        assert ('something=3.0=0', ) == bar_lock_set.package_specs_for_current_platform

        # be sure we can save
        lock_file.save()

        reloaded = ProjectLockFile.load_for_directory(dirname)

        assert ('something=3.0=0', ) == _get_lock_set(reloaded, 'bar').package_specs_for_current_platform
        assert _get_lock_set(reloaded, 'foo').disabled

    with_directory_contents({DEFAULT_PROJECT_LOCK_FILENAME: """
locking_enabled: false
"""}, check_file)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os
import pytest

from anaconda_project.internal.makedirs import makedirs_ok_if_exists
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents


def test_makedirs_ok_if_exists():
    def do_test(dirname):
        dir1 = os.path.join(dirname, "foo")
        dir2 = os.path.join(dir1, "bar")
        dir3 = os.path.join(dir2, "baz")
        assert os.path.isdir(dirname)
        assert not os.path.isdir(dir1)
        assert not os.path.isdir(dir2)
        assert not os.path.isdir(dir3)

        makedirs_ok_if_exists(dir3)

        assert os.path.isdir(dir1)
        assert os.path.isdir(dir2)
        assert os.path.isdir(dir3)

    with_directory_contents(dict(), do_test)


def test_makedirs_ok_if_exists_fails_for_another_reason(monkeypatch):
    def do_test(dirname):
        def mock_mkdir_fails(path, mode):
            raise IOError("not EEXIST")

        monkeypatch.setattr("os.mkdir", mock_mkdir_fails)

        with pytest.raises(IOError) as excinfo:
            makedirs_ok_if_exists("foo/bar/baz")
        assert 'not EEXIST' in repr(excinfo.value)

    with_directory_contents(dict(), do_test)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function, division, unicode_literals

import collections
import errno
import json
import os
import platform
import re
import shutil
import sys
import tempfile
import yaml

from anaconda_project.internal import streaming_popen
from anaconda_project.internal.directory_contains import subdirectory_relative_to_directory
from anaconda_project.internal.py2_compat import is_string, is_dict

CONDA_EXE = os.environ.get("CONDA_EXE", "conda")


class CondaError(Exception):
    """General Conda error."""
    def __init__(self, message, json=None):
        super(CondaError, self).__init__(message)
        self.json = json


class CondaEnvExistsError(CondaError):
    """Conda environment already exists."""

    pass


class CondaEnvMissingError(CondaError):
    """Conda environment missing."""

    pass


# this function exists so we can monkeypatch it in tests
def _get_conda_command(extra_args):
    # just use whatever conda is on the path
    cmd_list = [CONDA_EXE]
    cmd_list.extend(extra_args)
    return cmd_list


# This is obviously ridiculous, we'll work to
# find a better way (at least in newer versions
# of conda).
def _platform_hacked_conda_code(platform, bits):
    return """import conda
try:
    # this is conda 4.2 and 4.3

    # fix whether default channels have msys
    import conda.base.constants
    from conda.base.constants import DEFAULT_CHANNELS_UNIX, DEFAULT_CHANNELS_WIN
    if "{platform}" == 'win':
        corrected_channels = DEFAULT_CHANNELS_WIN
    else:
        corrected_channels = DEFAULT_CHANNELS_UNIX

    setattr(conda.base.constants, 'DEFAULT_CHANNELS', corrected_channels)

    from conda.base.context import Context

    class KapselHackedContext(Context):
        @property
        def subdir(self):
            return "{platform}-{bits}"

        @property
        def bits(self):
            return {bits}

    setattr(conda.base.context.context, "__class__", KapselHackedContext)
except ImportError:
    # this is conda 4.1
    import conda.config

    setattr(conda.config, "platform", "{platform}")
    setattr(conda.config, "bits", "{bits}")
    setattr(conda.config, "subdir", "{platform}-{bits}")

    # fix up the default urls
    msys_url = 'https://repo.continuum.io/pkgs/msys2'
    if "{platform}" == "win":
        if msys_url not in conda.config.defaults_:
            conda.config.defaults_.append(msys_url)
    else:
        if msys_url in conda.config.defaults_:
            conda.config.defaults_.remove(msys_url)


import conda.cli
import sys

sys.argv[0] = "conda"
sys.exit(conda.cli.main())
""".format(platform=platform, bits=bits).strip() + "\n"


def _get_platform_hacked_conda_command(extra_args, platform):
    """Get conda command and a string representing it in error messages."""
    if platform == current_platform() or platform is None:
        cmd_list = _get_conda_command(extra_args)
        return (cmd_list, " ".join(cmd_list))
    else:
        (platform_name, bits) = platform.split("-")

        conda_code = _platform_hacked_conda_code(platform_name, bits)

        # this has to run with the python from the root env,
        # so the conda modules will be found.
        root_prefix = _get_root_prefix()
        root_python = None
        for location in (('bin', 'python'), ('python.exe', ), ('Scripts', 'python.exe'), ('Library', 'bin',
                                                                                          'python.exe')):
            candidate = os.path.join(root_prefix, *location)
            if os.path.isfile(candidate):
                root_python = candidate
                break
        assert root_python is not None

        cmd_list = [root_python, '-c', conda_code]
        cmd_list.extend(extra_args)
        return (cmd_list, " ".join(["conda"] + cmd_list[3:]))


def _call_conda(extra_args, json_mode=False, platform=None, stdout_callback=None, stderr_callback=None):
    assert len(extra_args) > 0  # we deref extra_args[0] below

    (cmd_list, command_in_errors) = _get_platform_hacked_conda_command(extra_args, platform=platform)

    try:
        (p, stdout_lines, stderr_lines) = streaming_popen.popen(cmd_list,
                                                                stdout_callback=stdout_callback,
                                                                stderr_callback=stderr_callback)
    except OSError as e:
        raise CondaError("failed to run: %r: %r" % (command_in_errors, repr(e)))
    errstr = "".join(stderr_lines)
    if p.returncode != 0:
        parsed = None
        message = errstr
        if json_mode:
            try:
                out = "".join(stdout_lines)
                parsed = json.loads(out)
                if parsed is not None and isinstance(parsed, dict):
                    # some versions of conda do 'error' and others
                    # both 'error' and 'message' and they appear to
                    # be the same.
                    for field in ('message', 'error'):
                        if field in parsed:
                            message = parsed[field]
                            break
            except Exception:
                pass

        raise CondaError('%s: %s' % (command_in_errors, message), json=parsed)
    elif errstr != '' and stderr_callback is None:
        # this is a sort of fallback because not all of our code
        # passes in a callback yet.
        for line in stderr_lines:
            print("%s %s: %s" % ("conda", extra_args[0], line.strip()), file=sys.stderr)

    return "".join(stdout_lines)


def _call_and_parse_json(extra_args, platform=None):
    out = _call_conda(extra_args, json_mode=True, platform=platform)
    try:
        return json.loads(out)
    except ValueError as e:
        raise CondaError('Invalid JSON from conda: %s' % str(e))


def info(platform=None):
    """Return a dictionary with configuration information.

    No guarantee is made about which keys exist.  Therefore this function
    should only be used for testing and debugging.
    """
    return _call_and_parse_json(['info', '--json'], platform=platform)


def get_env_vars(env_prefix):
    """Return a dictionary of environment variables.

    These are Conda Environment variables that have been
    set using `conda env config vars`.
    """

    try:
        return _call_and_parse_json(['env', 'config', 'vars', 'list', '-p', env_prefix, '--json'])
    except CondaError as e:  # pragma: no cover
        # conda env config was introduced in version 4.8
        if 'invalid choice' not in str(e).lower():
            raise  # pragma: no cover
        else:
            return {}


def resolve_env_to_prefix(name_or_prefix):
    """Convert an env name or path into a canonical prefix path.

    Returns:
        Absolute path of prefix or None if it isn't found.
    """
    if os.path.isabs(name_or_prefix):
        return name_or_prefix

    json = info()
    root_prefix = json.get('root_prefix', None)
    if name_or_prefix == 'root':
        return root_prefix

    envs = json.get('envs', [])
    for prefix in envs:
        if os.path.basename(prefix) == name_or_prefix:
            return prefix
    return None


_cached_root_prefix = None


def _get_root_prefix():
    global _cached_root_prefix

    if _cached_root_prefix is None:
        _cached_root_prefix = resolve_env_to_prefix('root')
    return _cached_root_prefix


def create(prefix, pkgs=None, channels=(), stdout_callback=None, stderr_callback=None):
    """Create an environment either by name or path with a specified set of packages."""
    if os.path.exists(prefix):
        raise CondaEnvExistsError('Conda environment [%s] already exists' % prefix)

    cmd_list = ['create', '--yes', '--prefix', prefix]

    for channel in channels:
        cmd_list.extend(['--channel', channel])

    cmd_list.extend(pkgs)
    _call_conda(cmd_list, stdout_callback=stdout_callback, stderr_callback=stderr_callback)


def clone(prefix, source, stdout_callback=None, stderr_callback=None):
    """Clone a pre-existing env."""
    if not os.path.exists(source):
        raise CondaEnvMissingError('Conda environment [%s] does not exist to clone.' % source)

    cmd_list = ['create', '-p', prefix, '--clone', source]
    _call_conda(cmd_list, stdout_callback=stdout_callback, stderr_callback=stderr_callback)
    # If someone is using the .readonly flag-file approach, the clone command is going to copy
    # that. So we need to remove it if we find it in the new, copied environment.
    readonly_file = os.path.join(prefix, '.readonly')
    if os.path.exists(readonly_file):
        os.unlink(readonly_file)


def install(prefix, pkgs=None, channels=(), stdout_callback=None, stderr_callback=None):
    """Install packages into an environment either by name or path with a specified set of packages."""
    if not pkgs or not isinstance(pkgs, (list, tuple)):
        raise TypeError('must specify a list of one or more packages to install into existing environment, not %r',
                        pkgs)

    cmd_list = ['install', '--yes']
    cmd_list.extend(['--prefix', prefix])

    for channel in channels:
        cmd_list.extend(['--channel', channel])

    cmd_list.extend(pkgs)
    _call_conda(cmd_list, stdout_callback=stdout_callback, stderr_callback=stderr_callback)


def remove(prefix, pkgs=None, stdout_callback=None, stderr_callback=None):
    """Remove packages from an environment either by name or path."""
    if not pkgs or not isinstance(pkgs, (list, tuple)):
        raise TypeError('must specify a list of one or more packages to remove from existing environment')

    cmd_list = ['remove', '--yes']
    cmd_list.extend(['--prefix', prefix])

    cmd_list.extend(pkgs)
    _call_conda(cmd_list, stdout_callback=stdout_callback, stderr_callback=stderr_callback)


def _parse_dist(dist):
    # the "dist" is the basename of a package inside
    # conda-meta, like "numpy-1.10.4-py34_1"
    pieces = dist.rsplit('-', 2)
    if len(pieces) == 3:
        return tuple(pieces)
    else:
        return None


def installed(prefix):
    """Get a dict of package names to (name, version, build) tuples."""
    meta_dir = os.path.join(prefix, 'conda-meta')
    try:
        full_names = set(fn[:-5] for fn in os.listdir(meta_dir) if fn.endswith('.json'))
    except OSError as e:
        if e.errno == errno.ENOENT:
            full_names = set()
        else:
            raise CondaError(str(e))
    result = dict()
    for full_name in full_names:
        pieces = _parse_dist(full_name)
        if pieces is not None:
            result[pieces[0]] = pieces
    return result


def installed_pip(prefix):
    cmd_list = ['env', 'export', '-p', prefix]
    stdout = _call_conda(cmd_list)
    parsed = yaml.safe_load(stdout)
    dependencies = parsed.get('dependencies', [])
    for dep in dependencies:
        if is_dict(dep):
            return dep.get('pip', [])


def resolve_dependencies(pkgs, channels=(), platform=None):
    """Resolve packages into a full transitive list of (name, version, build) tuples."""
    if not pkgs or not isinstance(pkgs, (list, tuple)):
        raise TypeError('must specify a list of one or more packages to install into existing environment, not %r',
                        pkgs)

    # even with --dry-run, conda wants to create the prefix,
    # so we ensure it's somewhere out of the way.
    prefix = tempfile.mkdtemp(prefix="_anaconda_project_resolve_")

    # conda 4.1 (and possibly other versions) will complain
    # if the directory already exists. An evil attacker
    # on a multiuser system could replace this with a file
    # after we remove it, and then conda's mkdir would fail.
    os.rmdir(prefix)

    cmd_list = ['create', '--yes', '--quiet', '--json', '--dry-run', '--prefix', prefix]

    for channel in channels:
        cmd_list.extend(['--channel', channel])

    cmd_list.extend(pkgs)
    try:
        parsed = _call_and_parse_json(cmd_list, platform=platform)
    finally:
        try:
            if os.path.isdir(prefix):
                shutil.rmtree(prefix)
        except Exception:
            pass

    results = []
    actions = parsed.get('actions', [])
    # old conda gives us one dict, newer a list of dicts
    if isinstance(actions, dict):
        actions = [actions]

    for action in actions:
        if isinstance(action, dict):
            links = action.get('LINK', [])
            for link in links:
                found = None
                # 4.1 conda gives us a string like
                # 'python-3.6.0-0 2' and 4.3 gives us a
                # dict with the fields already decomposed.
                if isinstance(link, dict):
                    name = link.get('name', None)
                    version = link.get('version', None)
                    build_string = link.get('build_string', None)
                    if name is not None and \
                       version is not None and \
                       build_string is not None:
                        found = (name, version, build_string)
                elif is_string(link):
                    # we have a string like 'python-3.6.0-0 2'
                    pieces = link.split()
                    if len(pieces) > 0:
                        # 'found' can be None if we didn't understand the string
                        found = _parse_dist(pieces[0])

                if found is not None:
                    results.append(found)

    if len(results) == 0:
        raise CondaError("Could not understand JSON from Conda, could be a problem with this Conda version.",
                         json=parsed)

    return results


def _contains_conda_meta(path):
    conda_meta = os.path.join(path, "conda-meta")
    return os.path.isdir(conda_meta)


def _is_conda_bindir_unix(path):
    if path.endswith("/"):
        path = path[:-1]
    if not path.endswith("/bin"):
        return False
    possible_prefix = os.path.dirname(path)
    return _contains_conda_meta(possible_prefix)


def _path_endswith_windows(path, suffix):
    if path.endswith("\\") or path.endswith("/"):
        path = path[:-1]
    replaced = suffix.replace("\\", "/")
    return path.endswith("\\" + suffix) or \
        path.endswith("/" + suffix) or \
        path.endswith("\\" + replaced) or \
        path.endswith("/" + replaced)


def _is_conda_bindir_windows(path):
    # on Windows there are three conda binary locations:
    #   - the prefix itself (contains python.exe)
    #   - prefix\Library\bin
    #   - prefix\Scripts
    if path.endswith("\\") or path.endswith("/"):
        path = path[:-1]
    if _contains_conda_meta(path):
        return True
    elif _path_endswith_windows(path, "Library\\bin"):
        possible_prefix = os.path.dirname(os.path.dirname(path))
        return _contains_conda_meta(possible_prefix)
    elif _path_endswith_windows(path, "Scripts"):
        possible_prefix = os.path.dirname(path)
        return _contains_conda_meta(possible_prefix)
    else:
        return False


def _windows_bindirs(prefix):
    # activate.bat in conda-env does it in this order, [ prefix, Scripts, Library\bin ]
    dirs = [prefix]
    for item in ("Scripts", "Library\\bin"):
        dirs.append(os.path.join(prefix, item))
    return dirs


def _unix_bindirs(prefix):
    return [os.path.join(prefix, "bin")]


def _set_conda_env_in_path(path, prefix, bindirs_func, is_bindir_func):
    elements = path.split(os.pathsep)
    new_elements = []
    if prefix is not None:
        new_elements = bindirs_func(prefix)
    for element in elements:
        if element != "" and not is_bindir_func(element):
            new_elements.append(element)

    return os.pathsep.join(new_elements)


def _set_conda_env_in_path_unix(path, prefix):
    return _set_conda_env_in_path(path, prefix, _unix_bindirs, _is_conda_bindir_unix)


def _set_conda_env_in_path_windows(path, prefix):
    return _set_conda_env_in_path(path, prefix, _windows_bindirs, _is_conda_bindir_windows)


def set_conda_env_in_path(path, prefix):
    """Remove any existing conda envs in the given path string, then add the given one.

    Args:
        path (str): value of the PATH environment variable
        prefix (str): the environment prefix, or None to remove all conda bindirs
    Returns:
        the new PATH value
    """
    if platform.system() == 'Windows':
        return _set_conda_env_in_path_windows(path, prefix)
    else:
        return _set_conda_env_in_path_unix(path, prefix)


ParsedSpec = collections.namedtuple(
    'ParsedSpec', ['name', 'conda_constraint', 'pip_constraint', 'exact_version', 'exact_build_string'])

# this is copied from conda
_spec_pat = re.compile(
    r'''
(?P<name>[^=<>!\s]+)               # package name
\s*                                # ignore spaces
(
  (?P<cc>=[^=<>!]+(=[^=<>!]+)?)    # conda constraint
  |
  (?P<pc>[=<>!]{1,2}.+)            # new (pip-style) constraint(s)
)?
$                                  # end-of-line
''', re.VERBOSE)

_conda_constraint_pat = re.compile('=(?P<version>[^=<>!]+)(?P<build>=[^=<>!]+)?', re.VERBOSE)


def parse_spec(spec):
    """Parse a package name and version spec as conda would.

    Returns:
       ``ParsedSpec`` or None on failure
    """
    if not is_string(spec):
        raise TypeError("Expected a string not %r" % spec)

    m = _spec_pat.match(spec)
    if m is None:
        return None
    name = m.group('name').lower()
    pip_constraint = m.group('pc')
    if pip_constraint is not None:
        pip_constraint = pip_constraint.replace(' ', '')

    conda_constraint = m.group('cc')

    exact_version = None
    exact_build_string = None
    if conda_constraint is not None:
        m = _conda_constraint_pat.match(conda_constraint)
        assert m is not None
        exact_version = m.group('version')
        for special in ('|', '*', ','):
            if special in exact_version:
                exact_version = None
                break
        if exact_version is not None:
            exact_build_string = m.group('build')
            if exact_build_string is not None:
                assert exact_build_string[0] == '='
                exact_build_string = exact_build_string[1:]

    return ParsedSpec(name=name,
                      conda_constraint=conda_constraint,
                      pip_constraint=pip_constraint,
                      exact_version=exact_version,
                      exact_build_string=exact_build_string)


# these are in order of preference. On pre-4.1.4 Windows,
# CONDA_PREFIX and CONDA_ENV_PATH aren't set, so we get to
# CONDA_DEFAULT_ENV.
_all_prefix_variables = ('CONDA_PREFIX', 'CONDA_ENV_PATH', 'CONDA_DEFAULT_ENV')


def conda_prefix_variable():
    # conda 4.1.4 and higher sets CONDA_PREFIX to the full prefix,
    # and CONDA_DEFAULT_ENV to the env name only, cross-platform.

    # Pre-4.1.4, on Windows, activate.bat never sets
    # CONDA_ENV_PATH but sets CONDA_DEFAULT_ENV to the full
    # path to the environment.

    # Pre-4.1.4, on Unix, activate script sets CONDA_ENV_PATH
    # to the full path, and sets CONDA_DEFAULT_ENV to either
    # just the env name or the full path.

    # if we're in a conda environment, then use CONDA_PREFIX if it
    # was set by conda, otherwise use CONDA_ENV_PATH if set,
    # otherwise use CONDA_DEFAULT_ENV if set.
    for name in _all_prefix_variables:
        if name in os.environ:
            return name

    # if we aren't in a conda environment, just hope we have a
    # newer conda...
    return 'CONDA_PREFIX'


def environ_get_prefix(environ):
    for name in _all_prefix_variables:
        if name in environ:
            return environ.get(name)
    return None


def environ_delete_prefix_variables(environ):
    for name in _all_prefix_variables:
        if name in environ:
            del environ[name]


_envs_dirs = None
_root_dir = None


def environ_set_prefix(environ, prefix, varname=conda_prefix_variable()):
    prefix = os.path.normpath(prefix)
    environ[varname] = prefix
    if varname != 'CONDA_DEFAULT_ENV':
        # This case matters on both Unix and Windows
        # with conda >= 4.1.4 since requirement.env_var
        # is CONDA_PREFIX, and matters on Unix only pre-4.1.4
        # when requirement.env_var is CONDA_ENV_PATH.
        global _envs_dirs
        global _root_dir
        if _envs_dirs is None:
            i = info()
            _envs_dirs = [os.path.normpath(d) for d in i.get('envs_dirs', [])]
            _root_dir = os.path.normpath(i.get('root_prefix'))
        if prefix == _root_dir:
            name = 'root'
        else:
            for d in _envs_dirs:
                name = subdirectory_relative_to_directory(prefix, d)
                if name != prefix:
                    break
        environ['CONDA_DEFAULT_ENV'] = name


# This isn't all (e.g. leaves out arm, power).  it's sort of "all
# that people typically publish for"
default_platforms = ('linux-64', 'osx-64', 'win-64')
assert tuple(sorted(default_platforms)) == default_platforms

# osx-32 isn't in here since it isn't much used
default_platforms_plus_32_bit = ('linux-32', 'linux-64', 'osx-64', 'win-32', 'win-64')
assert tuple(sorted(default_platforms_plus_32_bit)) == default_platforms_plus_32_bit

_non_x86_linux_machines = {'armv6l', 'armv7l', 'ppc64le'}

# this list will get outdated, unfortunately.
_known_platforms = tuple(
    sorted(list(default_platforms_plus_32_bit) + ['osx-32'] + [("linux-%s" % m) for m in _non_x86_linux_machines]))

known_platform_names = ('linux', 'osx', 'win')
assert tuple(sorted(known_platform_names)) == known_platform_names

unix_platform_names = ('linux', 'osx')
assert tuple(sorted(unix_platform_names)) == unix_platform_names

_known_platform_groups = dict()

# Fill in the 'linux', 'osx', 'win' groups
for name in known_platform_names:
    result = []
    for p in default_platforms_plus_32_bit:
        if p.startswith(name):
            result.append(p)
    _known_platform_groups[name] = tuple(result)
    assert tuple(sorted(_known_platform_groups[name])) == _known_platform_groups[name]


# fill in the 'unix' group
def _known_unix_platforms():
    result = []
    for unix_name in unix_platform_names:
        for p in default_platforms_plus_32_bit:
            if p.startswith(unix_name):
                result.append(p)
    return tuple(result)


_known_platform_groups['unix'] = _known_unix_platforms()
assert tuple(sorted(_known_platform_groups['unix'])) == _known_platform_groups['unix']

# fill in the 'all' group
_known_platform_groups['all'] = default_platforms_plus_32_bit

# this isn't just _known_platform_groups.keys() because we want to be
# in order from most to least general
_known_platform_groups_keys = ('all', 'unix') + known_platform_names
assert set(_known_platform_groups_keys) == set(_known_platform_groups.keys())


def current_platform():
    m = platform.machine()
    if m in _non_x86_linux_machines:
        return 'linux-%s' % m
    else:
        _platform_map = {
            'linux2': 'linux',
            'linux': 'linux',
            'darwin': 'osx',
            'win32': 'win',
        }
        p = _platform_map.get(sys.platform, 'unknown')
        return '%s-%d' % (p, (8 * tuple.__itemsize__))


_default_platforms_with_current = tuple(sorted(list(set(default_platforms + (current_platform(), )))))


def default_platforms_with_current():
    return _default_platforms_with_current


def parse_platform(platform):
    """Split platform into OS name and architecture."""
    assert '-' in platform
    # platforms can have multiple hyphens e.g. linux-cos5-64 Our
    # goal here is to separate the general name from the
    # bit-width.
    pieces = platform.rsplit("-", 1)
    return (pieces[0], pieces[1])


def validate_platform_list(platforms):
    """Split platform list into known, unknown, and invalid platforms.

    Also, sort the list into canonical order.

    We return a tuple, the second list in the tuple
    is a subset of the first, and indicates platforms
    we don't know about. These may create a warning.
    The third list is not in the other two and indicates
    unusably-invalid platform names.

    Returns:
       Tuple of known platforms and unknown platforms.
    """
    result = set()
    unknown = set()
    invalid = set()
    for p in platforms:
        if '-' not in p:
            invalid.add(p)
        else:
            result.add(p)
            if p not in _known_platforms:
                unknown.add(p)

    # unknown platforms aren't necessarily an error, we just
    # don't do anything smart with them.
    return (sort_platform_list(result), sort_platform_list(unknown), sort_platform_list(invalid))


def sort_platform_list(platforms):
    """Sort platform list (including "grouping" names) from more to less general."""
    remaining = set(platforms)
    result = []
    for known in (_known_platform_groups_keys + _known_platforms):
        if known in remaining:
            result.append(known)
            remaining.remove(known)

    result = result + sorted(list(remaining))

    return result



# -*- coding: utf-8 -*-
# Multipart encoder from requests-toolbelt
# https://github.com/sigmavirus24/requests-toolbelt/blob/ef6e0bc12fe406e0bd730b8150deb5267e18fae9/requests_toolbelt/multipart/encoder.py

# Copyright 2014 Ian Cordasco, Cory Benfield

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""

requests_toolbelt.multipart.encoder
===================================

This holds all of the implementation details of the MultipartEncoder

"""
import contextlib
import io
import os
from uuid import uuid4

import anaconda_project.internal.test.fields as fields


class MultipartEncoder(object):
    """

    The ``MultipartEncoder`` oject is a generic interface to the engine that
    will create a ``multipart/form-data`` body for you.

    The basic usage is:

    .. code-block:: python

        import requests
        from requests_toolbelt import MultipartEncoder

        encoder = MultipartEncoder({'field': 'value',
                                    'other_field', 'other_value'})
        r = requests.post('https://httpbin.org/post', data=encoder,
                          headers={'Content-Type': encoder.content_type})

    If you do not need to take advantage of streaming the post body, you can
    also do:

    .. code-block:: python

        r = requests.post('https://httpbin.org/post',
                          data=encoder.to_string(),
                          headers={'Content-Type': encoder.content_type})

    If you want the encoder to use a specific order, you can use an
    OrderedDict or more simply, a list of tuples:

    .. code-block:: python

        encoder = MultipartEncoder([('field', 'value'),
                                    ('other_field', 'other_value')])

    .. versionchanged:: 0.4.0

    You can also provide tuples as part values as you would provide them to
    requests' ``files`` parameter.

    .. code-block:: python

        encoder = MultipartEncoder({
            'field': ('file_name', b'{"a": "b"}', 'application/json',
                      {'X-My-Header': 'my-value'})
        ])

    .. warning::

        This object will end up directly in :mod:`httplib`. Currently,
        :mod:`httplib` has a hard-coded read size of **8192 bytes**. This
        means that it will loop until the file has been read and your upload
        could take a while. This is **not** a bug in requests. A feature is
        being considered for this object to allow you, the user, to specify
        what size should be returned on a read. If you have opinions on this,
        please weigh in on `this issue`_.

    .. _this issue:
        https://github.com/sigmavirus24/requests-toolbelt/issues/75

    """
    def __init__(self, fields, boundary=None, encoding='utf-8'):
        #: Boundary value either passed in by the user or created
        self.boundary_value = boundary or uuid4().hex

        # Computed boundary
        self.boundary = '--{0}'.format(self.boundary_value)

        #: Encoding of the data being passed in
        self.encoding = encoding

        # Pre-encoded boundary
        self._encoded_boundary = b''.join(
            [encode_with(self.boundary, self.encoding),
             encode_with('\r\n', self.encoding)])

        #: Fields provided by the user
        self.fields = fields

        #: Whether or not the encoder is finished
        self.finished = False

        #: Pre-computed parts of the upload
        self.parts = []

        # Pre-computed parts iterator
        self._iter_parts = iter([])

        # The part we're currently working with
        self._current_part = None

        # Cached computation of the body's length
        self._len = None

        # Our buffer
        self._buffer = CustomBytesIO(encoding=encoding)

        # Pre-compute each part's headers
        self._prepare_parts()

        # Load boundary into buffer
        self._write_boundary()

    @property
    def len(self):
        """Length of the multipart/form-data body.

        requests will first attempt to get the length of the body by calling
        ``len(body)`` and then by checking for the ``len`` attribute.

        On 32-bit systems, the ``__len__`` method cannot return anything
        larger than an integer (in C) can hold. If the total size of the body
        is even slightly larger than 4GB users will see an OverflowError. This
        manifested itself in `bug #80`_.

        As such, we now calculate the length lazily as a property.

        .. _bug #80:
            https://github.com/sigmavirus24/requests-toolbelt/issues/80
        """
        # If _len isn't already calculated, calculate, return, and set it
        return self._len or self._calculate_length()

    def __repr__(self):
        return '<MultipartEncoder: {0!r}>'.format(self.fields)

    def _calculate_length(self):
        """
        This uses the parts to calculate the length of the body.

        This returns the calculated length so __len__ can be lazy.
        """
        boundary_len = len(self.boundary)  # Length of --{boundary}
        # boundary length + header length + body length + len('\r\n') * 2
        self._len = sum((boundary_len + total_len(p) + 4) for p in self.parts) + boundary_len + 4
        return self._len

    def _calculate_load_amount(self, read_size):
        """This calculates how many bytes need to be added to the buffer.

        When a consumer read's ``x`` from the buffer, there are two cases to
        satisfy:

            1. Enough data in the buffer to return the requested amount
            2. Not enough data

        This function uses the amount of unread bytes in the buffer and
        determines how much the Encoder has to load before it can return the
        requested amount of bytes.

        :param int read_size: the number of bytes the consumer requests
        :returns: int -- the number of bytes that must be loaded into the
            buffer before the read can be satisfied. This will be strictly
            non-negative
        """
        amount = read_size - total_len(self._buffer)
        return amount if amount > 0 else 0

    def _load(self, amount):
        """Load ``amount`` number of bytes into the buffer."""
        self._buffer.smart_truncate()
        part = self._current_part or self._next_part()
        while amount == -1 or amount > 0:
            written = 0
            if not part.bytes_left_to_write():
                written += self._write(b'\r\n')
                written += self._write_boundary()
                part = self._next_part()

            if not part:
                written += self._write_closing_boundary()
                self.finished = True
                break

            written += part.write_to(self._buffer, amount)

            if amount != -1:
                amount -= written

    def _next_part(self):
        try:
            p = self._current_part = next(self._iter_parts)
        except StopIteration:
            p = None
        return p

    def _iter_fields(self):
        _fields = self.fields
        if hasattr(self.fields, 'items'):
            _fields = list(self.fields.items())
        for k, v in _fields:
            file_name = None
            file_type = None
            file_headers = None
            if isinstance(v, (list, tuple)):
                if len(v) == 2:
                    file_name, file_pointer = v
                elif len(v) == 3:
                    file_name, file_pointer, file_type = v
                else:
                    file_name, file_pointer, file_type, file_headers = v
            else:
                file_pointer = v

            field = fields.RequestField(name=k, data=file_pointer, filename=file_name, headers=file_headers)
            field.make_multipart(content_type=file_type)
            yield field

    def _prepare_parts(self):
        """This uses the fields provided by the user and creates Part objects.

        It populates the `parts` attribute and uses that to create a
        generator for iteration.
        """
        enc = self.encoding
        self.parts = [Part.from_field(f, enc) for f in self._iter_fields()]
        self._iter_parts = iter(self.parts)

    def _write(self, bytes_to_write):
        """Write the bytes to the end of the buffer.

        :param bytes bytes_to_write: byte-string (or bytearray) to append to
            the buffer
        :returns: int -- the number of bytes written
        """
        return self._buffer.append(bytes_to_write)

    def _write_boundary(self):
        """Write the boundary to the end of the buffer."""
        return self._write(self._encoded_boundary)

    def _write_closing_boundary(self):
        """Write the bytes necessary to finish a multipart/form-data body."""
        with reset(self._buffer):
            self._buffer.seek(-2, 2)
            self._buffer.write(b'--\r\n')
        return 2

    def _write_headers(self, headers):
        """Write the current part's headers to the buffer."""
        return self._write(encode_with(headers, self.encoding))

    @property
    def content_type(self):
        return str('multipart/form-data; boundary={0}'.format(self.boundary_value))

    def to_string(self):
        return self.read()

    def read(self, size=-1):
        """Read data from the streaming encoder.

        :param int size: (optional), If provided, ``read`` will return exactly
            that many bytes. If it is not provided, it will return the
            remaining bytes.
        :returns: bytes
        """
        if self.finished:
            return self._buffer.read(size)

        bytes_to_load = size
        if bytes_to_load != -1 and bytes_to_load is not None:
            bytes_to_load = self._calculate_load_amount(int(size))

        self._load(bytes_to_load)
        return self._buffer.read(size)


def IDENTITY(monitor):
    return monitor


class MultipartEncoderMonitor(object):
    """
    An object used to monitor the progress of a :class:`MultipartEncoder`.

    The :class:`MultipartEncoder` should only be responsible for preparing and
    streaming the data. For anyone who wishes to monitor it, they shouldn't be
    using that instance to manage that as well. Using this class, they can
    monitor an encoder and register a callback. The callback receives the
    instance of the monitor.

    To use this monitor, you construct your :class:`MultipartEncoder` as you
    normally would.

    .. code-block:: python

        from requests_toolbelt import (MultipartEncoder,
                                       MultipartEncoderMonitor)
        import requests

        def callback(encoder, bytes_read):
            # Do something with this information
            pass

        m = MultipartEncoder(fields={'field0': 'value0'})
        monitor = MultipartEncoderMonitor(m, callback)
        headers = {'Content-Type': montior.content_type}
        r = requests.post('https://httpbin.org/post', data=monitor,
                          headers=headers)

    Alternatively, if your use case is very simple, you can use the following
    pattern.

    .. code-block:: python

        from requests_toolbelt import MultipartEncoderMonitor
        import requests

        def callback(encoder, bytes_read):
            # Do something with this information
            pass

        monitor = MultipartEncoderMonitor.from_fields(
            fields={'field0': 'value0'}, callback
            )
        headers = {'Content-Type': montior.content_type}
        r = requests.post('https://httpbin.org/post', data=monitor,
                          headers=headers)

    """
    def __init__(self, encoder, callback=None):
        #: Instance of the :class:`MultipartEncoder` being monitored
        self.encoder = encoder

        #: Optionally function to call after a read
        self.callback = callback or IDENTITY

        #: Number of bytes already read from the :class:`MultipartEncoder`
        #: instance
        self.bytes_read = 0

        #: Avoid the same problem in bug #80
        self.len = self.encoder.len

    @classmethod
    def from_fields(cls, fields, boundary=None, encoding='utf-8', callback=None):
        encoder = MultipartEncoder(fields, boundary, encoding)
        return cls(encoder, callback)

    @property
    def content_type(self):
        return self.encoder.content_type

    def to_string(self):
        return self.read()

    def read(self, size=-1):
        string = self.encoder.read(size)
        self.bytes_read += len(string)
        self.callback(self)
        return string


def encode_with(string, encoding):
    """Encoding ``string`` with ``encoding`` if necessary.

    :param str string: If string is a bytes object, it will not encode it.
        Otherwise, this function will encode it with the provided encoding.
    :param str encoding: The encoding with which to encode string.
    :returns: encoded bytes object
    """
    if not (string is None or isinstance(string, bytes)):
        return string.encode(encoding)
    return string


def readable_data(data, encoding):
    """Coerce the data to an object with a ``read`` method."""
    if hasattr(data, 'read'):
        return data

    return CustomBytesIO(data, encoding)


def total_len(o):
    if hasattr(o, '__len__'):
        return len(o)

    if hasattr(o, 'len'):
        return o.len

    if hasattr(o, 'fileno'):
        try:
            fileno = o.fileno()
        except io.UnsupportedOperation:
            pass
        else:
            return os.fstat(fileno).st_size

    if hasattr(o, 'getvalue'):
        # e.g. BytesIO, cStringIO.StringIO
        return len(o.getvalue())


@contextlib.contextmanager
def reset(buffer):
    """Keep track of the buffer's current position and write to the end.

    This is a context manager meant to be used when adding data to the buffer.
    It eliminates the need for every function to be concerned with the
    position of the cursor in the buffer.
    """
    original_position = buffer.tell()
    buffer.seek(0, 2)
    yield
    buffer.seek(original_position, 0)


def coerce_data(data, encoding):
    """Ensure that every object's __len__ behaves uniformly."""
    if not isinstance(data, CustomBytesIO):
        if hasattr(data, 'getvalue'):
            return CustomBytesIO(data.getvalue(), encoding)

        if hasattr(data, 'fileno'):
            return FileWrapper(data)

        if not hasattr(data, 'read'):
            return CustomBytesIO(data, encoding)

    return data


def to_list(fields):
    if hasattr(fields, 'items'):
        return list(fields.items())
    return list(fields)


class Part(object):
    def __init__(self, headers, body):
        self.headers = headers
        self.body = body
        self.headers_unread = True
        self.len = len(self.headers) + total_len(self.body)

    @classmethod
    def from_field(cls, field, encoding):
        """Create a part from a Request Field generated by urllib3."""
        headers = encode_with(field.render_headers(), encoding)
        body = coerce_data(field.data, encoding)
        return cls(headers, body)

    def bytes_left_to_write(self):
        """Determine if there are bytes left to write.

        :returns: bool -- ``True`` if there are bytes left to write, otherwise
            ``False``
        """
        to_read = 0
        if self.headers_unread:
            to_read += len(self.headers)

        return (to_read + total_len(self.body)) > 0

    def write_to(self, buffer, size):
        """Write the requested amount of bytes to the buffer provided.

        The number of bytes written may exceed size on the first read since we
        load the headers ambitiously.

        :param CustomBytesIO buffer: buffer we want to write bytes to
        :param int size: number of bytes requested to be written to the buffer
        :returns: int -- number of bytes actually written
        """
        written = 0
        if self.headers_unread:
            written += buffer.append(self.headers)
            self.headers_unread = False

        while total_len(self.body) > 0 and (size == -1 or written < size):
            amount_to_read = size
            if size != -1:
                amount_to_read = size - written
            written += buffer.append(self.body.read(amount_to_read))

        return written


class CustomBytesIO(io.BytesIO):
    def __init__(self, buffer=None, encoding='utf-8'):
        buffer = encode_with(buffer, encoding)
        super(CustomBytesIO, self).__init__(buffer)

    def _get_end(self):
        current_pos = self.tell()
        self.seek(0, 2)
        length = self.tell()
        self.seek(current_pos, 0)
        return length

    @property
    def len(self):
        length = self._get_end()
        return length - self.tell()

    def append(self, bytes):
        with reset(self):
            written = self.write(bytes)
        return written

    def smart_truncate(self):
        to_be_read = total_len(self)
        already_read = self._get_end() - to_be_read

        if already_read >= to_be_read:
            old_bytes = self.read()
            self.seek(0, 0)
            self.truncate()
            self.write(old_bytes)
            self.seek(0, 0)  # We want to be at the beginning


class FileWrapper(object):
    def __init__(self, file_object):
        self.fd = file_object

    @property
    def len(self):
        return total_len(self.fd) - self.fd.tell()

    def read(self, length=-1):
        return self.fd.read(length)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

from anaconda_project.project_commands import CommandExecInfo

import os
import platform
import pytest


def test_execvpe_with_shell_on_unix(monkeypatch):
    if platform.system() == 'Windows':
        return

    executed = {}

    def mock_execvpe(file, args, env):
        executed['file'] = file
        executed['args'] = args
        executed['env'] = env

    monkeypatch.setattr('os.execvpe', mock_execvpe)
    info = CommandExecInfo(cwd=os.getcwd(), args=['foo bar'], shell=True, env=dict(FOO='bar'))

    info.execvpe()

    assert executed['file'] == '/bin/sh'
    assert executed['args'] == ['/bin/sh', '-c', 'foo bar']
    assert executed['env'] == dict(FOO='bar')


def test_execvpe_with_shell_on_windows(monkeypatch):
    def mock_platform_system():
        return 'Windows'

    monkeypatch.setattr('platform.system', mock_platform_system)

    executed = {}

    def mock_popen(args, env, cwd, shell):
        executed['args'] = args
        executed['env'] = env
        executed['cwd'] = cwd
        executed['shell'] = shell

        class MockProcess(object):
            def wait(self):
                return 0

        return MockProcess()

    monkeypatch.setattr('subprocess.Popen', mock_popen)

    info = CommandExecInfo(cwd='/somewhere', args=['foo bar'], shell=True, env=dict(FOO='bar'))
    with pytest.raises(SystemExit) as excinfo:
        info.execvpe()
    assert excinfo.value.code == 0

    assert executed['args'] == 'foo bar'
    assert executed['shell'] is True
    assert executed['env'] == dict(FOO='bar')
    assert executed['cwd'] == '/somewhere'



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Project file loading and manipulation."""
from __future__ import absolute_import

import os

from anaconda_project.yaml_file import YamlFile
from anaconda_project.env_spec import EnvSpec
import anaconda_project.internal.conda_api as conda_api

# these are in the order we'll use them if multiple are present
possible_project_file_names = ("anaconda-project.yml", "anaconda-project.yaml", "kapsel.yml", "kapsel.yaml")

DEFAULT_PROJECT_FILENAME = possible_project_file_names[0]


def _empty_default_env_spec():
    return (EnvSpec(name="default", channels=[], conda_packages=()), )


class ProjectFile(YamlFile):
    """Represents the ``anaconda-project.yml`` file which describes the project across machines/users.

    State that's specific to a machine/user/checkout/deployment
    should instead be in ``LocalStateFile``.  ``ProjectFile``
    would normally be checked in to source control or otherwise
    act as a shared resource.

    Be careful with creating your own instance of this class,
    because you have to think about when other code might load or
    save in a way that conflicts with your loads and saves.

    """

    template = '''
# This is an Anaconda project file.
#
# Here you can describe your project and how to run it.
# Use `anaconda-project run` to run the project.
# The file is in YAML format, please see http://www.yaml.org/start.html for more.
#

#
# Set the 'name' key to name your project
#
name:

#
# Set the 'icon' key to give your project an icon
#
icon:

#
# Set a one-sentence-or-so 'description' key with project details
#
description:

#
# In the commands section, list your runnable scripts, notebooks, and other code.
# Use `anaconda-project add-command` to add commands.
#
commands: {}

#
# In the variables section, list any environment variables your code depends on.
# Use `anaconda-project add-variable` to add variables.
#
variables: {}

#
# In the services section, list any services that should be
# available before your code runs.
# Use `anaconda-project add-service` to add services.
#
services: {}

#
# In the downloads section, list any URLs to download to local files
# before your code runs.
# Use `anaconda-project add-download` to add downloads.
#
downloads: {}

#
# In the packages section, list any packages that must be installed
# before your code runs.
# Use `anaconda-project add-packages` to add packages.
#
packages: []

#
# In the channels section, list any Conda channel URLs to be searched
# for packages.
#
# For example,
#
# channels:
#    - mychannel
#
channels: []

#
# In the platforms section, list platforms the project should work on
# Examples: "linux-64", "osx-64", "win-64"
# Use `anaconda-project add-platforms` to add platforms.
#
platforms: []

#
# You can define multiple, named environment specs.
# Each inherits any global packages or channels,
# but can have its own unique ones also.
# Use `anaconda-project add-env-spec` to add environment specs.
#
env_specs: {}
'''

    @classmethod
    def load_for_directory(cls, directory, default_env_specs_func=_empty_default_env_spec, scan_parents=True):
        """Load the project file from the given directory, even if it doesn't exist.

        If the directory has no project file, and the project file
        cannot be found in any parent directory, the loaded
        ``ProjectFile`` will be empty. It won't actually be
        created on disk unless you call ``save()``.

        If the file has syntax problems, this sets the
        ``corrupted`` and ``corrupted_error_message`` properties,
        and attempts to modify the file will raise an
        exception. If the project file has semantic problems, they
        are not detected by this class but are reported by the
        ``Project`` class.

        Args:
            directory (str): path to the project directory
            default_env_specs_func (function makes list of EnvSpec): if file is created, use these
            scan_parents (bool): if True search for anaconda-project.yml file in parent directories
                                 If one is found change the directory_path to its location.

        Returns:
            a new ``ProjectFile``

        """
        current_dir = directory
        while current_dir != os.path.realpath(os.path.dirname(current_dir)):
            for name in possible_project_file_names:
                path = os.path.join(current_dir, name)
                if os.path.isfile(path):
                    return ProjectFile(path)

            if scan_parents:
                current_dir = os.path.dirname(os.path.abspath(current_dir))
                continue
            else:
                break

        # No file was found, create a new one
        return ProjectFile(os.path.join(directory, DEFAULT_PROJECT_FILENAME), default_env_specs_func)

    def __init__(self, filename, default_env_specs_func=_empty_default_env_spec):
        """Construct a ``ProjectFile`` with the given filename and requirement registry.

        It's easier to use ``ProjectFile.load_for_directory()`` in most cases.

        If the file has syntax problems, this sets the
        ``corrupted`` and ``corrupted_error_message`` properties,
        and attempts to modify the file will raise an
        exception. If the project file has semantic problems, they
        are not detected by this class but are reported by the
        ``Project`` class.

        Args:
            filename (str): path to the project file
        """
        self._default_env_specs_func = default_env_specs_func
        self.project_dir = os.path.dirname(filename)
        super(ProjectFile, self).__init__(filename)

    def _fill_default_content(self, as_json):
        as_json['name'] = os.path.basename(os.path.dirname(self.filename))
        as_json['platforms'].extend(conda_api.default_platforms_with_current())

        assert self._default_env_specs_func is not None
        default_env_specs = self._default_env_specs_func()
        assert default_env_specs is not None
        for env_spec in default_env_specs:
            as_json['env_specs'][env_spec.name] = env_spec.to_json()

        if len(default_env_specs) == 1:
            # if there's only one env spec, keep it for name/description
            # and put the packages and channels up in the global sections
            spec_name = next(iter(as_json['env_specs']))
            spec_json = as_json['env_specs'][spec_name]

            def move_list_elements(src, dest):
                # we want to preserve the dest list object with comments
                del dest[:]
                dest.extend(src)
                del src[:]

            if 'packages' in spec_json:
                move_list_elements(spec_json['packages'], as_json['packages'])
            if 'channels' in spec_json:
                move_list_elements(spec_json['channels'], as_json['channels'])
            if 'platforms' in spec_json:
                move_list_elements(spec_json['platforms'], as_json['platforms'])
